#!/usr/bin/python
# -*- coding: utf-8 -*-
import collections
import datetime
import getopt
import glob
import os
import re
import socket
import sys
import syslog
import time


def main():
    """
    Parse command line arguments.
    Note: RHEL6 ships with Python 2.6 which has no 'argparse' module.
    """
    try:
        opts, args = getopt.getopt(sys.argv[1:], 'H:L:P:m')
    except getopt.GetoptError:
        usage(sys.argv[0])
        sys.exit(2)

    hours = 1
    location = '/var/log/httpd'
    file_path = 'access_log'
    mds = False
    for o, a in opts:
        if o == '-H':
            try:
                hours = int(a)
                if (hours <= 0):
                    usage(sys.argv[0])
                    sys.exit(2)
            except ValueError:
                usage(sys.argv[0])
                sys.exit(2)
        elif o == '-L':
            location = a
        elif o == '-P':
            file_path = a
        elif o == '-m':
            mds = True
        else:
            usage(sys.argv[0])
            sys.exit(2)

    parse_file(os.path.join(location, file_path), hours, mds)


def usage(program):
    print("Usage: {0} [options]\n\n".format(program),
          "Parse httpd log file and collect basic statitics.\n\n",
          "Optional arguments:\n",
          " -H [hours] time span in hours to process (default: 1)\n",
          " -L [location] log file location (default: /var/log/httpd)\n",
          " -P [path] log file path (default: access_log)\n",
          " -m send the message to MDS agent (default: False)")


def parse_file(logfile, hours, mds):
    """
    Parse the httpd logfile and collect the following statistics:
        - Total accesses in the past hour(s)
        - Unique visited hosts in the past hour(s)

    Try to parse the rotated log file under the following two conditions:
        1. The current log file is empty.
        2. The first log entry in the current file is newer than cutoff time.
    """
    rotated = False
    now = datetime.datetime.now()
    cutoff = now - datetime.timedelta(hours=hours)
    access_count = 0
    hosts = set()
    with open(logfile) as f:
        line_num = 0
        for line in f.readlines():
            loginfo = parse_line(line)
            line_num += 1
            hosts.add(loginfo.host)
            if loginfo.access_time >= cutoff:
                access_count += 1
                if line_num == 1:
                    rotated = True
        if line_num == 0:
            rotated = True

    if rotated:
        nextfile = next_logfile(logfile)
        if nextfile:
            with open(nextfile) as f:
                for line in f.readlines():
                    loginfo = parse_line(line)
                    hosts.add(loginfo.host)
                if loginfo.access_time >= cutoff:
                    access_count += 1

    timespan = '1 hour'
    if hours > 1:
        timespan = '{0} hours'.format(hours)

    syslog.syslog(syslog.LOG_INFO,
                  "Total access in the last {0}: {1}"
                  .format(timespan, access_count))
    syslog.syslog(syslog.LOG_INFO,
                  "Unique hosts in the last {0}: {1}"
                  .format(timespan, len(hosts)))

    if (mds):
        write_to_mds(now, access_count, len(hosts), hours)


def next_logfile(logfile):
    """
    Find the next log file among the potential rotated log files
    """
    logfiles = sorted(glob.glob(logfile + '*'), key=os.path.getmtime)
    if (len(logfiles) < 2):
        return None
    else:
        return logfiles[1]


def parse_line(line):
    """
    Parse a single line/entry in the httpd log file
    """
    pattern = re.compile(r'(.+?)\s+.+?\s+.+?\s+\[(.+?)\]')
    m = pattern.search(line)
    access_time = parse_time(m.group(2))
    host = m.group(1)
    return collections.namedtuple('LogInfo',
                                  ['host', 'access_time'])(host, access_time)


def parse_time(s):
    """
    Parse Apache httpd time format *without* timezone, i.e. assuming the
    datetime is in local timezone set by the OS.  Later when we call
    time.mktime() to convert the datetime into Unix epoch time, the function
    has the same assumption.
    """
    month_map = {'Jan': 1, 'Feb': 2, 'Mar': 3, 'Apr': 4, 'May': 5, 'Jun': 6,
                 'Jul': 7, 'Aug': 8,  'Sep': 9, 'Oct': 10, 'Nov': 11, 'Dec': 12}

    obj = datetime.datetime(year=int(s[7:11]), month=month_map[s[3:6]],
                            day=int(s[0:2]), hour=int(s[12:14]),
                            minute=int(s[15:17]), second=int(s[18:20]))
    return obj


def write_to_mds(access_time, access_count, unique_hosts, hours):
    """
    Write the statistics to MDS agent using dynamic schema json. Log the
    error message to syslog and exit with error on failure. No retry attemps.
    """
    source = 'httpd'
    msg_id = 0
    schema_id = 0
    schema = ('[["AccessTime","FT_TIME"],'
              '["AccessCount","FT_INT32"],'
              '["UniqueHosts","FT_INT32"],'
              '["TimeSpanHrs","FT_INT32"]]')

    msg = '["{0}",{1},{2},{3},[[{4},0],{5},{6},{7}]]'.format(
        source, msg_id, schema_id, schema,
        int(time.mktime(access_time.timetuple())),
        access_count, unique_hosts, hours)

    server_address = '/var/run/mdsd/default_djson.socket'
    sock = socket.socket(socket.AF_UNIX, socket.SOCK_STREAM)
    try:
        sock.connect(server_address)
        sock.send(str(len(msg)))
        sock.send('\n')
        sock.send(msg)

        ack = sock.recv(1024)  # 1024 should be enough for mdsd ack msg
        resp = int(ack.strip().split(':', 2)[1])
        if resp != 0:
            syslog.syslog(syslog.LOG_ERR,
                          "Fail to send the message to mdsd." +
                          "Response code: {0}".format(resp))
            sys.exit(3)
    finally:
        sock.close()


if __name__ == "__main__":
    main()
